#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>

#include "life.h"
#include "lookup.h"
#include "util.h"

/* Private Function Prototypes */

cage_type*  birth_cage(int32 x, int32 y);
void        clear_cage_list(cage_type** start);
void        get_block_text(uint16 block, char buf[8][9], int32 xoffset, int32 yoffset);
cage_type*  find_cage(int32 x, int32 y);
cage_type*  find_rattle_birth(int32 x, int32 y);
void        kill_cage(cage_type* cage);
uint32      population_count(uint16 nw, uint16 ne, uint16 sw, uint16 se);
void        rattle_cage(cage_type* cage);
void        tranquilize_cage(cage_type* cage);
void        update_onscreen_flag(cage_type* cage);

/* Defined in the GUI code */
void draw_life_block(uint16 block, int32 xstart, int32 ystart, boolean full_update);

/* Constants */

static uint8 bitcount[0x100] = {
    0, 1, 1, 2, 1, 2, 2, 3, 1, 2, 2, 3, 2, 3, 3, 4,
    1, 2, 2, 3, 2, 3, 3, 4, 2, 3, 3, 4, 3, 4, 4, 5,
    1, 2, 2, 3, 2, 3, 3, 4, 2, 3, 3, 4, 3, 4, 4, 5,
    2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6,
    1, 2, 2, 3, 2, 3, 3, 4, 2, 3, 3, 4, 3, 4, 4, 5,
    2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6,
    2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6,
    3, 4, 4, 5, 4, 5, 5, 6, 4, 5, 5, 6, 5, 6, 6, 7,
    1, 2, 2, 3, 2, 3, 3, 4, 2, 3, 3, 4, 3, 4, 4, 5,
    2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6,
    2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6,
    3, 4, 4, 5, 4, 5, 5, 6, 4, 5, 5, 6, 5, 6, 6, 7,
    2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6,
    3, 4, 4, 5, 4, 5, 5, 6, 4, 5, 5, 6, 5, 6, 6, 7,
    3, 4, 4, 5, 4, 5, 5, 6, 4, 5, 5, 6, 5, 6, 6, 7,
    4, 5, 5, 6, 5, 6, 6, 7, 5, 6, 6, 7, 6, 7, 7, 8
};

/* Public Globals */

cage_type*  active_cages;      /* doubly-linked list */
cage_type*  sleeping_cages;    /* doubly-linked list */
uint32      tick;
boolean     parity;            /* on an odd (1) or even (0) tick? */
uint32      population;

char        pattern_description[DESCRIPTION_MAX_LINES][DESCRIPTION_MAX_COLS+1];
int32       desc_num_lines;

/* Private Globals */

static cage_type*  dead_cages;   /* singly-linked list */
static cage_type*  cage_hash[XY_HASH_SIZE];
static uint32      sleeping_population;

static int32       viewport_xstart = -1, viewport_ystart = -1;
static int32       viewport_xend   = -1, viewport_yend   = -1;

/* Public Functions */

/* Completely clear the universe, emptying all cage lists, emptying the hash, resetting ticks to 0
 * and pattern_description to NULL.
 */
void clear_world(void)
{
    /* Clear all cage lists */
    clear_cage_list(&active_cages);
    clear_cage_list(&sleeping_cages);
    clear_cage_list(&dead_cages);

    /* Zero-out the hash */
    memset(cage_hash, 0, XY_HASH_SIZE * sizeof(cage_type*));

    /* Set ticks and population to 0 */
    tick = parity = population = sleeping_population = 0;

    /* Clear the description */
    desc_num_lines = 0;
}

/* Set, unset, or toggle the cell at the given XY coordinates.
 */
void draw_cell(int32 x, int32 y, draw_method_type draw_method)
{
    cage_type*  cage;
    int32       cagex, cagey;
    int32       xoffset, yoffset;
    uint16*     block;
    uint16      bitmask;

    /* Check if coordinates are out of bounds */
    if (x < 0 || x >= WORLD_SIZE || y < 0 || y >= WORLD_SIZE)
        return;

    /* Find the cage containing this cell, rattle it, birth it if necessary */
    cagex = x / CAGE_SIZE;
    cagey = y / CAGE_SIZE;
    cage = find_cage(cagex, cagey);
    if (cage)
        rattle_cage(cage);
    else {
        if (draw_method == DRAW_UNSET)
            return;
        else
            cage = birth_cage(cagex, cagey);
    }

    /* Find/rattle/birth all neighbors */
    if (cage->north)
        rattle_cage(cage->north);
    else {
        cage->north = find_rattle_birth(cagex, cagey-1);
        if (cage->north)
            cage->north->south = cage;
    }
    if (cage->south)
        rattle_cage(cage->south);
    else {
        cage->south = find_rattle_birth(cagex, cagey+1);
        if (cage->south)
            cage->south->north = cage;
    }
    if (cage->west)
        rattle_cage(cage->west);
    else {
        cage->west = find_rattle_birth(cagex-1, cagey);
        if (cage->west)
            cage->west->east = cage;
    }
    if (cage->east)
        rattle_cage(cage->east);
    else {
        cage->east = find_rattle_birth(cagex+1, cagey);
        if (cage->east)
            cage->east->west = cage;
    }
    if (cage->nw)
        rattle_cage(cage->nw);
    else {
        cage->nw = find_rattle_birth(cagex-1, cagey-1);
        if (cage->nw)
            cage->nw->se = cage;
    }
    if (cage->ne)
        rattle_cage(cage->ne);
    else {
        cage->ne = find_rattle_birth(cagex+1, cagey-1);
        if (cage->ne)
            cage->ne->sw = cage;
    }
    if (cage->sw)
        rattle_cage(cage->sw);
    else {
        cage->sw = find_rattle_birth(cagex-1, cagey+1);
        if (cage->sw)
            cage->sw->ne = cage;
    }
    if (cage->se)
        rattle_cage(cage->se);
    else {
        cage->se = find_rattle_birth(cagex+1, cagey+1);
        if (cage->se)
            cage->se->nw = cage;
    }

    /* Extract the uint16 representing the appropriate 4x4 block; get offsets relative
     * to the block.
     */
    xoffset = x % CAGE_SIZE;
    yoffset = y % CAGE_SIZE;
    if (yoffset < 4) {
        if (xoffset < 4)
            block = &(cage->bnw[parity]);
        else
            block = &(cage->bne[parity]);
    } else {
        if (xoffset < 4)
            block = &(cage->bsw[parity]);
        else
            block = &(cage->bse[parity]);
    }
    yoffset %= 4;
    xoffset %= 4;

    /* Set, unset, or toggle the appropriate bit */
    bitmask = 1 << (yoffset*4 + xoffset);
    if (draw_method == DRAW_SET) {
        if (!(*block & bitmask))
            population++;
        *block |= bitmask;
    }
    else {
        if (*block & bitmask)
            population--;
        *block &= ~bitmask;
    }
}

/* Load a pattern from the given file, returning true if successful.
 */
load_result_type load_pattern(const char* path)
{
    FILE*   f;
    char*   line;
    char*   desc;
    char*   p;
    double  version;
    int32   x, y, curx;

    /* Open the file */
    f = fopen(path, "r");
    if (!f)
        return ((errno == ENOENT) ? LOAD_NOT_FOUND : LOAD_SYS_ERROR);

    /* Make sure this is a proper .lif file. Check pattern-file version, make sure its at
     * or below PATTERNS_MAX_VERSION
     */
    line = dgets(f);
    if (!line || !STR_STARTS_WITH(line, "#Life"))
        return LOAD_INVALID_FORMAT;
    if (sscanf(line+5, "%lf", &version) != 1 || version > PATTERNS_MAX_VERSION)
        return LOAD_INVALID_FORMAT;
    free(line);

    /* The file appears to be legit, so clear the world */
    clear_world();

    /* Load the file */
    line = dgets(f);
    while (line) {
        if (STR_STARTS_WITH(line, "#D")) {
            /* Load a description line, unless we are at max lines */
            if (desc_num_lines < DESCRIPTION_MAX_LINES) {
                desc = line+2;
                trim_whitespace(desc);
                if (strlen(desc) > DESCRIPTION_MAX_COLS)
                    desc[DESCRIPTION_MAX_COLS] = '\0';
                strcpy(pattern_description[desc_num_lines++], desc);
            }
        } else if (STR_STARTS_WITH(line, "#P")) {
            /* Load a block of cell data */
            if (sscanf(line+2, "%d%d", &x, &y) != 2)
                x = y = 0;
            x += WORLD_SIZE / 2;
            y += WORLD_SIZE / 2;
            free(line);
            while ((line = dgets(f)) != NULL && !STR_STARTS_WITH(line, "#")) {
                /* Load this line of data, unless we are out of bounds or the line is empty */
                if (y < WORLD_SIZE && line[0] != '\0') {
                    for (p=line, curx=x; *p; p++, curx++) {
                        if (curx < WORLD_SIZE && *p == '*')
                            draw_cell(curx, y, DRAW_SET);
                    }
                    y++;
                }
                free(line);
            }
            continue;
        }

        free(line);
        line = dgets(f);
    }

    fclose(f);
    return LOAD_SUCCESS;
}

/* Calculate the next Life generation, and update the pixmap via calls to the external function
 * draw_life_block.
 */
void next_tick(void)
{
    cage_type*  cage;
    cage_type*  next = NULL;
    int32       turns_stable;
    int32       t1, t2;
    int32       x, y;
    uint16      nw, ne, sw, se;
    uint16      neigh_nnw, neigh_nne, neigh_ssw, neigh_sse, neigh_wnw, neigh_wsw,
                neigh_ene, neigh_ese, neigh_nw, neigh_ne, neigh_sw, neigh_se;
    uint16      new_nw, new_ne, new_sw, new_se;

    t1 = parity;
    t2 = !parity;
    population = sleeping_population;

    for (cage=active_cages; cage; cage=next) {
        /* Store next-pointer in case we end up killing or tranqing this cage */
        next = cage->next;

        /* Extract the four 4x4 blocks in this cage, and any neighboring 4x4 blocks that
         * we'll need to know about.
         */
        neigh_nnw = neigh_nne = neigh_ssw = neigh_sse = neigh_wnw = neigh_wsw =
            neigh_ene = neigh_ese = neigh_nw = neigh_ne = neigh_sw = neigh_se = 0;
        nw = cage->bnw[t1];
        ne = cage->bne[t1];
        sw = cage->bsw[t1];
        se = cage->bse[t1];
        if (cage->north) {
            neigh_nnw = cage->north->bsw[t1];
            neigh_nne = cage->north->bse[t1];
        }
        if (cage->south) {
            neigh_ssw = cage->south->bnw[t1];
            neigh_sse = cage->south->bne[t1];
        }
        if (cage->west) {
            neigh_wnw = cage->west->bne[t1];
            neigh_wsw = cage->west->bse[t1];
        }
        if (cage->east) {
            neigh_ene = cage->east->bnw[t1];
            neigh_ese = cage->east->bsw[t1];
        }
        if (cage->nw)
            neigh_nw = cage->nw->bse[t1];
        if (cage->ne)
            neigh_ne = cage->ne->bsw[t1];
        if (cage->sw)
            neigh_sw = cage->sw->bne[t1];
        if (cage->se)
            neigh_se = cage->se->bnw[t1];

        /* Calculate the new 4x4 blocks for this cage */
        new_nw =
            ul_lookup[(nw & 0x0777) | (neigh_nnw & 0x7000) | (neigh_wnw & 0x0888) | (neigh_nw & 0x8000)]   |
            ur_lookup[(nw & 0x0EEE) | (neigh_nnw & 0xE000) | (ne & 0x0111)        | (neigh_nne & 0x1000)]  |
            ll_lookup[(nw & 0x7770) | (sw & 0x0007)        | (neigh_wnw & 0x8880) | (neigh_wsw & 0x0008)]  |
            lr_lookup[(nw & 0xEEE0) | (sw & 0x000E)        | (ne & 0x1110)        | (se & 0x0001)];
        new_ne =
            ul_lookup[(ne & 0x0777) | (neigh_nne & 0x7000) | (nw & 0x0888)        | (neigh_nnw & 0x8000)]  |
            ur_lookup[(ne & 0x0EEE) | (neigh_nne & 0xE000) | (neigh_ene & 0x0111) | (neigh_ne & 0x1000)]   |
            ll_lookup[(ne & 0x7770) | (se & 0x0007)        | (nw & 0x8880)        | (sw & 0x0008)]         |
            lr_lookup[(ne & 0xEEE0) | (se & 0x000E)        | (neigh_ene & 0x1110) | (neigh_ese & 0x0001)];
        new_sw =
            ul_lookup[(sw & 0x0777) | (nw & 0x7000)        | (neigh_wsw & 0x0888) | (neigh_wnw & 0x8000)]  |
            ur_lookup[(sw & 0x0EEE) | (nw & 0xE000)        | (se & 0x0111)        | (ne & 0x1000)]         |
            ll_lookup[(sw & 0x7770) | (neigh_ssw & 0x0007) | (neigh_wsw & 0x8880) | (neigh_sw & 0x0008)]   |
            lr_lookup[(sw & 0xEEE0) | (neigh_ssw & 0x000E) | (se & 0x1110)        | (neigh_sse & 0x0001)];
        new_se =
            ul_lookup[(se & 0x0777) | (ne & 0x7000)        | (sw & 0x0888)        | (nw & 0x8000)]         |
            ur_lookup[(se & 0x0EEE) | (ne & 0xE000)        | (neigh_ese & 0x0111) | (neigh_ene & 0x1000)]  |
            ll_lookup[(se & 0x7770) | (neigh_sse & 0x0007) | (sw & 0x8880)        | (neigh_ssw & 0x0008)]  |
            lr_lookup[(se & 0xEEE0) | (neigh_sse & 0x000E) | (neigh_ese & 0x1110) | (neigh_se & 0x0001)];

        /* Update population count */
        population += population_count(new_nw, new_ne, new_sw, new_se);

        /* Check if this cage has changed, and update TURNS_STABLE accordingly. If it's been stable
         * awhile, move it to sleeping_cages or dead_cages (if empty).
         */
        if (new_nw == nw && new_ne == ne && new_sw == sw && new_se == se) {
            turns_stable = TURNS_STABLE(cage);
            if (turns_stable == MAX_TURNS_STABLE) {
                if (!new_nw && !new_ne && !new_sw && !new_se)
                    kill_cage(cage);
                else
                    tranquilize_cage(cage);
                continue;        /* no need to apply new blocks, redraw, or rattle neighbors */
            }
            else {
                INC_TURNS_STABLE(cage);
                if (turns_stable)
                    continue;    /* no need to apply new blocks, redraw, or rattle neighbors */
                turns_stable++;
            }
        } else {
            RESET_TURNS_STABLE(cage);
            turns_stable = 0;
        }

        /* Apply the new blocks */
        cage->bnw[t2] = new_nw;
        cage->bne[t2] = new_ne;
        cage->bsw[t2] = new_sw;
        cage->bse[t2] = new_se;

        if (turns_stable)
            continue;  /* no need to redraw or rattle neighbors */

        /* Update the Life pixmap if necessary */
        if (IS_ONSCREEN(cage)) {
            x = cage->x * CAGE_SIZE;
            y = cage->y * CAGE_SIZE;
            if (new_nw != nw)
                draw_life_block(new_nw, x, y, FALSE);
            if (new_ne != ne)
                draw_life_block(new_ne, x + BLOCK_SIZE, y, FALSE);
            if (new_sw != sw)
                draw_life_block(new_sw, x, y + BLOCK_SIZE, FALSE);
            if (new_se != se)
                draw_life_block(new_se, x + BLOCK_SIZE, y + BLOCK_SIZE, FALSE);
        }

        /* Rattle/find/create neighbors */
        x = cage->x;
        y = cage->y;
        if ((new_nw & NORTH_EDGE_MASK) != (nw & NORTH_EDGE_MASK) ||
            (new_ne & NORTH_EDGE_MASK) != (ne & NORTH_EDGE_MASK)) {
            if (cage->north)
                rattle_cage(cage->north);
            else {
                cage->north = find_rattle_birth(x, y-1);
                if (cage->north)
                    cage->north->south = cage;
            }
        }
        if ((new_sw & SOUTH_EDGE_MASK) != (sw & SOUTH_EDGE_MASK) ||
            (new_se & SOUTH_EDGE_MASK) != (se & SOUTH_EDGE_MASK)) {
            if (cage->south)
                rattle_cage(cage->south);
            else {
                cage->south = find_rattle_birth(x, y+1);
                if (cage->south)
                    cage->south->north = cage;
            }
        }
        if ((new_nw & WEST_EDGE_MASK) != (nw & WEST_EDGE_MASK) ||
            (new_sw & WEST_EDGE_MASK) != (sw & WEST_EDGE_MASK)) {
            if (cage->west)
                rattle_cage(cage->west);
            else {
                cage->west = find_rattle_birth(x-1, y);
                if (cage->west)
                    cage->west->east = cage;
            }
        }
        if ((new_ne & EAST_EDGE_MASK) != (ne & EAST_EDGE_MASK) ||
            (new_se & EAST_EDGE_MASK) != (se & EAST_EDGE_MASK)) {
            if (cage->east)
                rattle_cage(cage->east);
            else {
                cage->east = find_rattle_birth(x+1, y);
                if (cage->east)
                    cage->east->west = cage;
            }
        }
        if ((new_nw & NW_CORNER_MASK) != (nw & NW_CORNER_MASK)) {
            if (cage->nw)
                rattle_cage(cage->nw);
            else {
                cage->nw = find_rattle_birth(x-1, y-1);
                if (cage->nw)
                    cage->nw->se = cage;
            }
        }
        if ((new_ne & NE_CORNER_MASK) != (ne & NE_CORNER_MASK)) {
            if (cage->ne)
                rattle_cage(cage->ne);
            else {
                cage->ne = find_rattle_birth(x+1, y-1);
                if (cage->ne)
                    cage->ne->sw = cage;
            }
        }
        if ((new_sw & SW_CORNER_MASK) != (sw & SW_CORNER_MASK)) {
            if (cage->sw)
                rattle_cage(cage->sw);
            else {
                cage->sw = find_rattle_birth(x-1, y+1);
                if (cage->sw)
                    cage->sw->ne = cage;
            }
        }
        if ((new_se & SE_CORNER_MASK) != (se & SE_CORNER_MASK)) {
            if (cage->se)
                rattle_cage(cage->se);
            else {
                cage->se = find_rattle_birth(x+1, y+1);
                if (cage->se)
                    cage->se->nw = cage;
            }
        }
    }

    tick++;
    parity = !parity;
}

/* Save the current pattern to the given file, returning true if successful, false if a system
 * error occurred in opening or writing the file.
 */
boolean save_pattern(const char* path)
{
    FILE*       f;
    cage_type*  c;
    char        chunk[8][9];
    boolean     processed_sleepers = FALSE;
    int32       startx, starty;
    int32       first_row, last_row, first_col;
    int32       x, y;
    int32       i;

    /* Open the file for writing and print the standard header */
    f = fopen(path, "w");
    if (!f)
        return FALSE;
    fprintf(f, "#Life %.2f\r\n", PATTERNS_MAX_VERSION);

    /* Output the description, if any */
    for (i=0; i < desc_num_lines; i++)
        fprintf(f, "#D %s\r\n", pattern_description[i]);

    /* Normal rules */
    fprintf(f, "#N\r\n");

    /* Output the cage data */
    for (y=0; y < 8; y++)
        chunk[y][8] = '\0';
    for (c = active_cages; ; c = c->next) {
        if (!c) {
            if (!processed_sleepers) {
                c = sleeping_cages;
                processed_sleepers = TRUE;
            }
            if (!c)
                break;
        }
        if (IS_EMPTY(c))
            continue;

        /* Get the starting coords for the cage, relative to center of world */
        startx = (c->x * CAGE_SIZE) - WORLD_SIZE/2;
        starty = (c->y * CAGE_SIZE) - WORLD_SIZE/2;

        /* Fill the chunk with the data from the four 4x4 blocks */
        get_block_text(c->bnw[parity], chunk, 0, 0);
        get_block_text(c->bne[parity], chunk, 4, 0);
        get_block_text(c->bsw[parity], chunk, 0, 4);
        get_block_text(c->bse[parity], chunk, 4, 4);

        /* Trim off empty rows, empty columns, and trailing empty cells */
        for (y=0; y < 8 && STR_EQUAL(chunk[y], "........"); y++, starty++) {}
        first_row = y;
        for (y=7; y > first_row && STR_EQUAL(chunk[y], "........"); y--) {}
        last_row = y;
        for (x=0; x < 8; x++, startx++) {
            for (y=0; y < 8 && chunk[y][x] == '.'; y++) {}
            if (y < 8)
                break;
        }
        first_col = x;
        for (y=0; y < 8; y++) {
            for (x=7; x > first_col && chunk[y][x] == '.'; x--) { }
            chunk[y][x+1] = '\0';
        }

        /* Output chunk header and data */
        fprintf(f, "#P %d %d\r\n", startx, starty);
        for (y=first_row; y <= last_row; y++)
            fprintf(f, "%s\r\n", chunk[y] + first_col);
    }

    /* Close file, returning FALSE if there was a write error */
    if (fclose(f) != 0)
        return FALSE;

    return TRUE;
}

/* Set the current Life viewport (in cell coordinates), which is used to determine when to call
 * draw_life_block() in next_tick().
 */
void set_viewport(int32 xstart, int32 ystart, int32 xend, int32 yend)
{
    cage_type*  cage;

    viewport_xstart = xstart;
    viewport_ystart = ystart;
    viewport_xend = xend;
    viewport_yend = yend;

    for (cage=active_cages; cage; cage=cage->next)
        update_onscreen_flag(cage);
    for (cage=sleeping_cages; cage; cage=cage->next)
        update_onscreen_flag(cage);
}

/* Private Functions */

/* Allocate the cage at the given x, y position, find its neighbors, and insert it into
 * the active cages list and cage hash. Return a pointer to the new cage. If x and/or
 * y is out of bounds, return NULL.
 */
cage_type* birth_cage(int32 x, int32 y)
{
    cage_type*  cage;
    uint16      hash;

    /* Return NULL if x and/or y is out of bounds */
    if (x < 0 || x >= WORLD_CAGES || y < 0 || y >= WORLD_CAGES)
        return NULL;

    /* Allocate, set X and Y */
    if (dead_cages) {
        cage = dead_cages;
        dead_cages = dead_cages->next;
        memset(cage, 0, sizeof(cage_type));
    } else
        cage = safe_calloc(1, sizeof(cage_type));
    cage->x = x;
    cage->y = y;

    /* Set whether it's onscreen or offscreen */
    update_onscreen_flag(cage);

    /* Find allocated neighbors */
    cage->north = find_cage(x, y-1);
    if (cage->north)
        cage->north->south = cage;
    cage->south = find_cage(x, y+1);
    if (cage->south)
        cage->south->north = cage;
    cage->west = find_cage(x-1, y);
    if (cage->west)
        cage->west->east = cage;
    cage->east = find_cage(x+1, y);
    if (cage->east)
        cage->east->west = cage;
    cage->nw = find_cage(x-1, y-1);
    if (cage->nw)
        cage->nw->se = cage;
    cage->ne = find_cage(x+1, y-1);
    if (cage->ne)
        cage->ne->sw = cage;
    cage->sw = find_cage(x-1, y+1);
    if (cage->sw)
        cage->sw->ne = cage;
    cage->se = find_cage(x+1, y+1);
    if (cage->se)
        cage->se->nw = cage;

    /* Insert into active cages list */
    cage->prev = NULL;
    cage->next = active_cages;
    if (active_cages)
        active_cages->prev = cage;
    active_cages = cage;

    /* Insert into cage hash */
    hash = XY_HASH(x,y);
    cage->hash_prev = NULL;
    cage->hash_next = cage_hash[hash];
    if (cage_hash[hash])
        cage_hash[hash]->hash_prev = cage;
    cage_hash[hash] = cage;

    return cage;
}

void clear_cage_list(cage_type** start)
{
    cage_type*  p;
    cage_type*  next;

    for (p=*start; p; p=next) {
        next = p->next;
        free(p);
    }
    *start = NULL;
}

/* Fill a quarter of the buffer buf with text representing the given 4x4 cell block, using the
 * given x/y offsets into the buffer.
 */
void get_block_text(uint16 block, char buf[8][9], int32 xoffset, int32 yoffset)
{
    int32  x, y;
    int32  bit;

    for (y=yoffset, bit=0; y < yoffset+4; y++) {
        for (x=xoffset; x < xoffset+4; x++, bit++)
            buf[y][x] = ((block & (1 << bit)) ? '*' : '.');
    }
}

/* Search the cage hash for the cage with the given XY coordinates. Return a pointer to
 * the cage, or NULL if it isn't allocated.
 */
cage_type* find_cage(int32 x, int32 y)
{
    uint16      hash;
    cage_type*  cage;

    if (x < 0 || x >= WORLD_CAGES || y < 0 || y >= WORLD_CAGES)
        return NULL;
    hash = XY_HASH(x,y);
    for (cage=cage_hash[hash];
         cage && !(cage->x == x && cage->y == y);
         cage = cage->hash_next) {}
    return cage;
}

/* Search the cage hash for the cage with the given XY coordinates. If found, rattle it
 * and return a pointer to it, otherwise birth and return a pointer to a new cage.
 */
cage_type* find_rattle_birth(int32 x, int32 y)
{
    cage_type* c;

    c = find_cage(x, y);
    if (c)
        rattle_cage(c);
    else
        c = birth_cage(x, y);
    return c;
}

/* Move a cage into the dead_cages pool, deregister it from the hash, and set neighboring pointers
 * to it to NULL.
 */
void kill_cage(cage_type* cage)
{
    /* Set neighboring pointers to self to NULL */
    if (cage->north)
        cage->north->south = NULL;
    if (cage->south)
        cage->south->north = NULL;
    if (cage->west)
        cage->west->east = NULL;
    if (cage->east)
        cage->east->west = NULL;
    if (cage->nw)
        cage->nw->se = NULL;
    if (cage->ne)
        cage->ne->sw = NULL;
    if (cage->sw)
        cage->sw->ne = NULL;
    if (cage->se)
        cage->se->nw = NULL;

    /* Move into dead_cages list */
    if (cage->prev)
        cage->prev->next = cage->next;
    else
        active_cages = cage->next;
    if (cage->next)
        cage->next->prev = cage->prev;
    cage->next = dead_cages;
    dead_cages = cage;

    /* Deregister from the cage hash */
    if (cage->hash_prev)
        cage->hash_prev->hash_next = cage->hash_next;
    else
        cage_hash[XY_HASH(cage->x, cage->y)] = cage->hash_next;
    if (cage->hash_next)
        cage->hash_next->hash_prev = cage->hash_prev;
}

/* Count the number of live cells in the cage with the given four blocks
 */
uint32 population_count(uint16 nw, uint16 ne, uint16 sw, uint16 se)
{
    return bitcount[FIRST_BYTE(nw)] + bitcount[SECOND_BYTE(nw)] +
           bitcount[FIRST_BYTE(ne)] + bitcount[SECOND_BYTE(ne)] +
           bitcount[FIRST_BYTE(sw)] + bitcount[SECOND_BYTE(sw)] +
           bitcount[FIRST_BYTE(se)] + bitcount[SECOND_BYTE(se)];
}

/* If a cage is sleeping, flag it as active and move it to the front of active_cages. Otherwise,
 * just make sure TURNS_STABLE isn't maxed out. Note: This function subtracts from
 * sleeping_population if necessary, but does not update the main population count.
 */
void rattle_cage(cage_type* cage)
{
    cage_type**  old_list;

    if (IS_ASLEEP(cage)) {
        SET_AWAKE(cage);
        old_list = &sleeping_cages;
        sleeping_population -= population_count(cage->bnw[0], cage->bne[0], cage->bsw[0],
                                                cage->bse[0]);
    }
    else {
        if (TURNS_STABLE(cage) == MAX_TURNS_STABLE)
            DEC_TURNS_STABLE(cage);
        return;
    }

    if (cage->prev)
        cage->prev->next = cage->next;
    else
        *old_list = cage->next;
    if (cage->next)
        cage->next->prev = cage->prev;
    cage->prev = NULL;
    cage->next = active_cages;
    if (active_cages)
        active_cages->prev = cage;
    active_cages = cage;
}

/* Flag a cage as asleep and move it to sleeping_cages. Note: This function adds to
 * sleeping_population if necessary, but does not update the main population count.
 */
void tranquilize_cage(cage_type* cage)
{
    SET_ASLEEP(cage);
    if (cage->prev)
        cage->prev->next = cage->next;
    else
        active_cages = cage->next;
    if (cage->next)
        cage->next->prev = cage->prev;
    cage->prev = NULL;
    cage->next = sleeping_cages;
    if (sleeping_cages)
        sleeping_cages->prev = cage;
    sleeping_cages = cage;
    sleeping_population += population_count(cage->bnw[0], cage->bne[0], cage->bsw[0],
                                            cage->bse[0]);
}

/* Set or clear the "onscreen" flag for the given cage, depending on viewport settings.
 */
void update_onscreen_flag(cage_type* cage)
{
    int32  x, y;

    x = cage->x * CAGE_SIZE;
    y = cage->y * CAGE_SIZE;
    if (x + CAGE_SIZE > viewport_xstart && x <= viewport_xend &&
        y + CAGE_SIZE > viewport_ystart && y <= viewport_yend) {
        SET_ONSCREEN(cage);
    }
    else
        SET_OFFSCREEN(cage);
}
