#ifndef LIFE_H
#define LIFE_H

#include "util.h"

/* Defines */

#define CAGE_SIZE               0x08
#define BLOCK_SIZE              0x04
#define WORLD_CAGES             0x2000
#define WORLD_SIZE              (CAGE_SIZE*WORLD_CAGES)

#define MAX_TURNS_STABLE        7

#define PATTERNS_MAX_VERSION    1.05
#define DESCRIPTION_MAX_LINES   22
#define DESCRIPTION_MAX_COLS    78

#define XY_HASH_BITS            7
#define XY_HASH_SIZE            (1 << (XY_HASH_BITS*2))
#define XY_HASH_MASK            ((1 << XY_HASH_BITS) - 1)

#define NORTH_EDGE_MASK         0x000F
#define SOUTH_EDGE_MASK         0xF000
#define WEST_EDGE_MASK          0x1111
#define EAST_EDGE_MASK          0x8888
#define NW_CORNER_MASK          0x0001
#define NE_CORNER_MASK          0x0008
#define SW_CORNER_MASK          0x1000
#define SE_CORNER_MASK          0x8000

#define TURNS_STABLE_MASK       0x07
#define SLEEPING_MASK           0x08
#define ONSCREEN_MASK           0x10

/* Macros */

#define IS_EMPTY(c)                 (!((c)->bnw[parity]) && !((c)->bne[parity]) && \
                                     !((c)->bsw[parity]) && !((c)->bse[parity]))
#define IS_ASLEEP(c)                ((c)->flags &   SLEEPING_MASK)
#define SET_ASLEEP(c)               ((c)->flags |=  SLEEPING_MASK)
#define SET_AWAKE(c)                ((c)->flags &= ~SLEEPING_MASK)
#define IS_ONSCREEN(c)              ((c)->flags &   ONSCREEN_MASK)
#define SET_ONSCREEN(c)             ((c)->flags |=  ONSCREEN_MASK)
#define SET_OFFSCREEN(c)            ((c)->flags &= ~ONSCREEN_MASK)
#define TURNS_STABLE(c)             ((c)->flags &   TURNS_STABLE_MASK)
#define RESET_TURNS_STABLE(c)       ((c)->flags &= ~TURNS_STABLE_MASK)
#define INC_TURNS_STABLE(c)         ((c)->flags++)
#define DEC_TURNS_STABLE(c)         ((c)->flags--)

#define FIRST_BYTE(x)               *((uint8*)(&x))
#define SECOND_BYTE(x)              *((uint8*)(&x) + 1)

#define XY_HASH(x,y)                ((((y) & XY_HASH_MASK) << XY_HASH_BITS) | ((x) & XY_HASH_MASK))

/* Enums */

typedef enum {DRAW_SET, DRAW_UNSET} draw_method_type;
typedef enum {DIR_NORTH, DIR_SOUTH, DIR_WEST, DIR_EAST,
              DIR_NW, DIR_NE, DIR_SW, DIR_SE,
              DIR_NONE}
        direction_type;

typedef enum {LOAD_SUCCESS, LOAD_NOT_FOUND, LOAD_SYS_ERROR, LOAD_INVALID_FORMAT} load_result_type;

/* Types */

/* An 8x8 cell block, with each 4x4 block (upper left, upper right, lower
 * left, lower right) stored as a uint16, and with pointers to neighbors.
 *
 * Flags:
 *   Bits 0-2: # of ticks since this cage last changed   (TURNS_STABLE)
 *   Bit    3: sleeping?    (IS_ASLEEP)
 *   Bit    4: onscreen?    (IS_ONSCREEN)
 */
typedef struct cage_struct {
    uint16  x, y;    /* start position by blocks */
    uint8   flags;
    uint16  bnw[2], bne[2], bsw[2], bse[2];    /* element 0 on even generations */
    struct cage_struct  *north, *south, *west, *east;
    struct cage_struct  *nw, *ne, *sw, *se;
    struct cage_struct  *next, *prev;
    struct cage_struct  *hash_next, *hash_prev;
} cage_type;

/* Externs */

extern cage_type*  active_cages;
extern cage_type*  sleeping_cages;
extern uint32      tick;
extern boolean     parity;    /* on an odd (1) or even (0) tick? */
extern uint32      population;

extern char        pattern_description[DESCRIPTION_MAX_LINES][DESCRIPTION_MAX_COLS+1];
extern int32       desc_num_lines;

/* Public Prototypes */

void              clear_world(void);
void              draw_cell(int32 x, int32 y, draw_method_type draw_method);
load_result_type  load_pattern(const char* path);
void              next_tick(void);
void              reset_ticker(void);
boolean           save_pattern(const char* path);
void              set_viewport(int32 xstart, int32 ystart, int32 xend, int32 yend);

#endif
